import * as nuxt_schema from 'nuxt/schema';
import * as valibot from 'valibot';
import { Input } from 'valibot';
import { UseScriptOptions } from '@unhead/schema';
import { UseScriptInput } from '@unhead/vue';

type NuxtUseScriptOptions<T = any> = Omit<UseScriptOptions<T>, 'trigger'> & {
    /**
     * The trigger to load the script:
     * - `onNuxtReady` - Load the script when Nuxt is ready.
     * - `manual` - Load the script manually by calling `$script.load()` or `$script.waitForLoad()`.
     * - `Promise` - Load the script when the promise resolves.
     */
    trigger?: UseScriptOptions<T>['trigger'] | 'onNuxtReady';
    /**
     * Should the script be bundled as an asset and loaded from your server. This is useful for improving the
     * performance by avoiding the extra DNS lookup and reducing the number of requests. It also
     * improves privacy by not sharing the user's IP address with third-party servers.
     */
    assetStrategy?: 'bundle';
    /**
     * A hook to run when a script does not exist and will be initialized for the first time.
     */
    beforeInit?: () => void;
};
type NuxtUseScriptInput = UseScriptInput;

interface CloudflareWebAnalyticsApi {
    __cfBeacon: {
        load: 'single';
        spa: boolean;
        token: string;
    };
    __cfRl?: unknown;
}
declare global {
    interface Window extends CloudflareWebAnalyticsApi {
    }
}
declare const CloudflareWebAnalyticsOptions: valibot.ObjectSchema<{
    /**
     * The Cloudflare Web Analytics token.
     *
     * Required when used for the first time.
     */
    token: valibot.StringSchema<string>;
    /**
     * Cloudflare Web Analytics enables measuring SPAs automatically by overriding the History API’s pushState function
     * and listening to the onpopstate. Hash-based router is not supported.
     *
     * @default true
     */
    spa: valibot.OptionalSchema<valibot.BooleanSchema<boolean>, undefined, boolean | undefined>;
}, undefined, {
    token: string;
    spa?: boolean | undefined;
}>;

interface ModuleOptions {
    /**
     * Register scripts globally.
     */
    register?: {
        cloudflareWebAnalytics?: Input<typeof CloudflareWebAnalyticsOptions>;
    };
    /**
     * Register scripts that should be loaded globally on all pages.
     */
    globals?: (NuxtUseScriptInput | [NuxtUseScriptInput, NuxtUseScriptOptions])[];
    /**
     * Override the static script options for specific scripts based on their provided `key` or `src`.
     */
    overrides?: {
        [key: string]: Pick<NuxtUseScriptOptions, 'assetStrategy'>;
    };
    /** Configure the way scripts assets are exposed */
    assets?: {
        /**
         * The baseURL where scripts files are served.
         * @default '/_scripts/'
         */
        prefix?: string;
        /** Currently scripts assets are exposed as public assets as part of the build. This will be configurable in future */
        strategy?: 'public';
    };
    /**
     * Whether the module is enabled.
     *
     * @default true
     */
    enabled: boolean;
    /**
     * Enables debug mode.
     *
     * @false false
     */
    debug: boolean;
}
interface ModuleHooks {
    /**
     * Transform a script before it's registered.
     */
    'scripts:transform': (ctx: {
        script: string;
        options: any;
    }) => Promise<void>;
}
declare const _default: nuxt_schema.NuxtModule<ModuleOptions>;

export { type ModuleHooks, type ModuleOptions, _default as default };
