import { useNuxt, useLogger, addDevServerHandler, createResolver, addTemplate, defineNuxtModule, addImportsDir, addImports, addPluginTemplate, addPlugin, addBuildPlugin } from '@nuxt/kit';
import { readPackageJSON } from 'pkg-types';
import { existsSync } from 'node:fs';
import { pathToFileURL } from 'node:url';
import { createUnplugin } from 'unplugin';
import { parseURL, parseQuery, joinURL, hasProtocol } from 'ufo';
import MagicString from 'magic-string';
import { walk } from 'estree-walker';
import fsp from 'node:fs/promises';
import { lazyEventHandler, eventHandler, createError } from 'h3';
import { fetch } from 'ofetch';
import { colors } from 'consola/utils';
import { defu } from 'defu';
import { join, relative } from 'pathe';
import { hash } from 'ohash';
import { createStorage } from 'unstorage';
import fsDriver from 'unstorage/drivers/fs-lite';

const DEVTOOLS_UI_ROUTE = "/__nuxt-scripts";
const DEVTOOLS_UI_LOCAL_PORT = 3030;
function setupDevToolsUI(options, resolve, nuxt = useNuxt()) {
  const clientPath = resolve("./client");
  const isProductionBuild = existsSync(clientPath);
  if (isProductionBuild) {
    nuxt.hook("vite:serverCreated", async (server) => {
      const sirv = await import('sirv').then((r) => r.default || r);
      server.middlewares.use(
        DEVTOOLS_UI_ROUTE,
        sirv(clientPath, { dev: true, single: true })
      );
    });
  } else {
    nuxt.hook("vite:extendConfig", (config) => {
      config.server = config.server || {};
      config.server.proxy = config.server.proxy || {};
      config.server.proxy[DEVTOOLS_UI_ROUTE] = {
        target: `http://localhost:${DEVTOOLS_UI_LOCAL_PORT}${DEVTOOLS_UI_ROUTE}`,
        changeOrigin: true,
        followRedirects: true,
        rewrite: (path) => path.replace(DEVTOOLS_UI_ROUTE, "")
      };
    });
  }
  nuxt.hook("devtools:customTabs", (tabs) => {
    tabs.push({
      // unique identifier
      name: "nuxt-scripts",
      // title to display in the tab
      title: "Scripts",
      // any icon from Iconify, or a URL to an image
      icon: "carbon:script",
      // iframe view
      view: {
        type: "iframe",
        src: DEVTOOLS_UI_ROUTE
      }
    });
  });
}

function NuxtScriptAssetBundlerTransformer(options) {
  return createUnplugin(() => {
    return {
      name: "nuxt:scripts:asset-bundler-transformer",
      enforce: "post",
      transformInclude(id) {
        const { pathname, search } = parseURL(decodeURIComponent(pathToFileURL(id).href));
        const { type } = parseQuery(search);
        if (pathname.endsWith(".vue") && (type === "script" || !search))
          return true;
        if (pathname.match(/\.((c|m)?j|t)sx?$/g))
          return true;
        if (pathname.includes("node_modules/@unhead") || pathname.includes("node_modules/vueuse"))
          return false;
        return false;
      },
      async transform(code, id) {
        if (!code.includes("useScript"))
          return;
        const ast = this.parse(code);
        const s = new MagicString(code);
        walk(ast, {
          enter(_node) {
            if (_node.type === "CallExpression" && _node.callee.type === "Identifier" && _node.callee?.name === "useScript") {
              const node = _node;
              let scriptKey;
              let scriptNode;
              if (node.arguments[0].type === "Literal") {
                scriptNode = node.arguments[0];
                scriptKey = scriptNode.value;
              } else if (node.arguments[0].type === "ObjectExpression") {
                const srcProperty = node.arguments[0].properties.find(
                  (p) => p.key?.name === "src" || p.key?.value === "src"
                );
                const keyProperty = node.arguments[0].properties.find(
                  (p) => p.key?.name === "key" || p.key?.value === "key"
                );
                scriptKey = keyProperty?.value?.value || srcProperty?.value;
                scriptNode = srcProperty?.value;
              }
              if (scriptNode) {
                const src = scriptNode.value;
                if (src) {
                  let hasAssetStrategy = false;
                  if (node.arguments[1]?.type === "ObjectExpression") {
                    const assetStrategyProperty = node.arguments[1]?.properties.find(
                      (p) => p.key?.name === "assetStrategy" || p.key?.value === "assetStrategy"
                    );
                    if (assetStrategyProperty) {
                      if (assetStrategyProperty?.value?.value !== "bundle")
                        return;
                      s.remove(assetStrategyProperty.start, assetStrategyProperty.end + 1);
                      hasAssetStrategy = true;
                    }
                  }
                  hasAssetStrategy = hasAssetStrategy || options.overrides?.[scriptKey]?.assetStrategy === "bundle";
                  if (hasAssetStrategy) {
                    const newSrc = options.resolveScript(src);
                    s.overwrite(scriptNode.start, scriptNode.end, `'${newSrc}'`);
                  }
                }
              }
            }
          }
        });
        if (s.hasChanged()) {
          return {
            code: s.toString(),
            map: s.generateMap({ includeContent: true, source: id })
          };
        }
      }
    };
  });
}

const logger = useLogger("@nuxt/scripts");

const ONE_YEAR_IN_SECONDS = 60 * 60 * 24 * 365;
function setupPublicAssetStrategy(options = {}) {
  const assetsBaseURL = options.prefix || "/_scripts";
  const nuxt = useNuxt();
  const renderedScriptSrc = /* @__PURE__ */ new Map();
  const storage = createStorage(fsDriver({
    base: "node_modules/.cache/nuxt/scripts"
  }));
  function normalizeScriptData(src) {
    if (hasProtocol(src, { acceptRelative: true })) {
      src = src.replace(/^\/\//, "https://");
      const url = parseURL(src);
      const file = [
        `${hash(url)}.js`
        // force an extension
      ].filter(Boolean).join("-");
      renderedScriptSrc.set(file, src);
      return joinURL(assetsBaseURL, file);
    }
    return src;
  }
  addDevServerHandler({
    route: assetsBaseURL,
    handler: lazyEventHandler(async () => {
      return eventHandler(async (event) => {
        const filename = event.path.slice(1);
        const url = renderedScriptSrc.get(event.path.slice(1));
        if (!url)
          throw createError({ statusCode: 404 });
        const key = `data:scripts:${filename}`;
        let res = await storage.getItemRaw(key);
        if (!res) {
          res = await fetch(url).then((r) => r.arrayBuffer()).then((r) => Buffer.from(r));
          await storage.setItemRaw(key, res);
        }
        return res;
      });
    })
  });
  if (nuxt.options.dev) {
    nuxt.options.routeRules ||= {};
    nuxt.options.routeRules[joinURL(assetsBaseURL, "**")] = {
      cache: {
        maxAge: ONE_YEAR_IN_SECONDS
      }
    };
  }
  nuxt.options.nitro.publicAssets ||= [];
  const cacheDir = join(nuxt.options.buildDir, "cache", "scripts");
  nuxt.options.nitro.publicAssets.push();
  nuxt.options.nitro = defu(nuxt.options.nitro, {
    publicAssets: [{
      dir: cacheDir,
      maxAge: ONE_YEAR_IN_SECONDS,
      baseURL: assetsBaseURL
    }],
    prerender: {
      ignore: [assetsBaseURL]
    }
  });
  nuxt.hook("nitro:init", async (nitro) => {
    if (nuxt.options.dev)
      return;
    nitro.hooks.hook("rollup:before", async () => {
      await fsp.rm(cacheDir, { recursive: true, force: true });
      await fsp.mkdir(cacheDir, { recursive: true });
      let banner = false;
      for (const [filename, url] of renderedScriptSrc) {
        const key = `data:scripts:${filename}`;
        let res = await storage.getItemRaw(key);
        if (!res) {
          if (!banner) {
            banner = true;
            logger.info("Downloading scripts...");
          }
          let encoding;
          let size = 0;
          res = await fetch(url).then((r) => {
            encoding = r.headers.get("content-encoding");
            size = Number(r.headers.get("content-length") / 1024);
            return r.arrayBuffer();
          }).then((r) => Buffer.from(r));
          logger.log(colors.gray(`  \u251C\u2500 ${url} \u2192 ${joinURL(assetsBaseURL, filename)} (${size.toFixed(2)} kB ${encoding})`));
          await storage.setItemRaw(key, res);
        }
        await fsp.writeFile(join(cacheDir, filename), res);
      }
      if (banner)
        logger.success("Scripts downloaded and cached.");
    });
  });
  return { normalizeScriptData };
}

function extendTypes(module, template) {
  const nuxt = useNuxt();
  const { resolve } = createResolver(import.meta.url);
  addTemplate({
    filename: `module/${module}.d.ts`,
    getContents: async () => {
      const typesPath = relative(resolve(nuxt.options.rootDir, nuxt.options.buildDir, "module"), resolve("runtime/types"));
      const s = await template({ typesPath });
      return `// Generated by ${module}
${s}
export {}
`;
    }
  });
  nuxt.hooks.hook("prepare:types", ({ references }) => {
    references.push({ path: resolve(nuxt.options.buildDir, `module/${module}.d.ts`) });
  });
}

const module = defineNuxtModule({
  meta: {
    name: "@nuxt/scripts",
    configKey: "scripts",
    compatibility: {
      nuxt: "^3.11.1",
      bridge: false
    }
  },
  defaults: {
    enabled: true,
    debug: false
  },
  async setup(config, nuxt) {
    const { resolve } = createResolver(import.meta.url);
    const { version, name } = await readPackageJSON(resolve("../package.json"));
    if (!config.enabled) {
      logger.debug("The module is disabled, skipping setup.");
      return;
    }
    nuxt.options.alias["#nuxt-scripts"] = resolve("./runtime/types");
    nuxt.options.runtimeConfig["nuxt-scripts"] = { version };
    addImportsDir([
      resolve("./runtime/composables")
    ]);
    const registry = [
      {
        name: "useScriptCloudflareTurnstile",
        from: resolve("./runtime/registry/cloudflare-turnstile")
      },
      {
        name: "useScriptCloudflareWebAnalytics",
        from: resolve("./runtime/registry/cloudflare-web-analytics")
      },
      {
        name: "useScriptConfetti",
        from: resolve("./runtime/registry/confetti")
      },
      {
        name: "useScriptFacebookPixel",
        from: resolve("./runtime/registry/facebook-pixel")
      },
      {
        name: "useScriptFathomAnalytics",
        from: resolve("./runtime/registry/fathom-analytics")
      },
      {
        name: "useScriptGoogleAnalytics",
        from: resolve("./runtime/registry/google-analytics")
      },
      {
        name: "useScriptGoogleTagManager",
        from: resolve("./runtime/registry/google-tag-manager")
      },
      {
        name: "useScriptHotjar",
        from: resolve("./runtime/registry/hotjar")
      },
      {
        name: "useScriptIntercom",
        from: resolve("./runtime/registry/intercom")
      },
      {
        name: "useScriptSegment",
        from: resolve("./runtime/registry/segment")
      }
    ].map((i) => {
      i.priority = -1;
      return i;
    });
    addImports(registry);
    const hasRegister = Object.keys(config.register || {}).length > 0;
    if (hasRegister) {
      addPluginTemplate({
        filename: "third-party.mjs",
        write: true,
        getContents() {
          const imports = ['import { defineNuxtPlugin } from "#imports";'];
          const inits = [];
          for (const [k, c] of Object.entries(config.register || {})) {
            const importPath = resolve(`./runtime/composables/${k}`);
            const exportName = k.substring(0, 1).toUpperCase() + k.substring(1);
            imports.unshift(`import { ${exportName} } from "${importPath}";`);
            inits.push(`${exportName}(${JSON.stringify(c)});`);
          }
          return [
            imports.join("\n"),
            "",
            "export default defineNuxtPlugin({",
            '  name: "nuxt-third-party",',
            "  setup() {",
            inits.map((i) => `    ${i}`).join("\n"),
            "  }",
            "})"
          ].join("\n");
        }
      });
    }
    if (config.globals?.length) {
      const template = addTemplate({
        filename: "modules/nuxt-scripts/plugin.client.mjs",
        getContents() {
          return `import { defineNuxtPlugin, useScript } from '#imports'
export default defineNuxtPlugin({
  setup() {
${config.globals?.map((g) => !Array.isArray(g) ? `    useScript("${g.toString()}")` : g.length === 2 ? `    useScript(${JSON.stringify(g[0])}, ${JSON.stringify(g[1])} })` : `    useScript(${JSON.stringify(g[0])})`).join("\n")}
  }
})`;
        }
      });
      addPlugin({
        src: template.dst,
        mode: "client"
      });
    }
    extendTypes(name, async () => {
      return `
declare module '#app' {
    interface NuxtApp {
      ${nuxt.options.dev ? `_scripts: (import('#nuxt-scripts').NuxtAppScript)[]` : ""}
    }
}
`;
    });
    const scriptMap = /* @__PURE__ */ new Map();
    const { normalizeScriptData } = setupPublicAssetStrategy(config.assets);
    addBuildPlugin(NuxtScriptAssetBundlerTransformer({
      overrides: config.overrides,
      resolveScript(src) {
        if (scriptMap.has(src))
          return scriptMap.get(src);
        const url = normalizeScriptData(src);
        scriptMap.set(src, url);
        return url;
      }
    }));
    if (nuxt.options.dev)
      setupDevToolsUI(config, resolve);
  }
});

export { module as default };
